#include <math.h>
#include "seispp.h"
namespace SEISPP {
/* this group of functions implement gap processing for time series
 * objects and their descendents called Three_Component_Seismograms.
 * They use indexing through the STL standard container called a
 * set.  The comparison function is defined in the seispp.h file
 * that allows interval indexing.  That is, we can look up a time
 * interval through an indexing algorithm that is an intrinsic part
 * of the generic set object.  This provides a fast lookup mechanism
 * to make this processing reasonably efficient.
 *
 * Author:  Gary L. Pavlis
 * Written:  May 2003
 */

//This is a standard function on Suns found in sunmath.  
// I've added this implementation here to get this to compile on linux
int nint(double x)
{
        double remainder;
        if(x==0.0)
                return(0);
        int i0 = (int)x;
        if(i0==0)
                remainder = x;
        else if(x>0.0)
                remainder = x - ((double)i0);
	else
		remainder = x-((double)i0);
        if(x>0)
        {
                if(remainder>0.5)
                        return(i0+1);
                else
                        return(i0);
        }
        else
        {
		if(remainder<-0.5)
                        return(i0-1);
                else
                        return(i0);
        }
}
// Returns true if the requested sample number of a gap or outside the
// range of the data
bool Basic_Time_Series::is_gap(int n0)
{
	if(n0<0 || n0>ns) return true;
	if(gaps.empty()) return false;
	// We use a window of 1 sample centered on sample time
	// This is should always work as long for processing where
	// we don't need to worry about overlapping waveforms with slipperly 
	// clocks
	Time_Window twin;
	double t=time(n0);
	twin.start = t - dt*0.5;
	twin.end = t +  dt*0.5;
	if(gaps.find(twin)==gaps.end()) 
		return false;
	else
		return true;
}
bool Basic_Time_Series::is_gap(double t)
{
	if(t<t0 || t>(t0+((double)(ns-1))*dt)) return true;
	if(gaps.empty())return false;
	Time_Window twin;
	twin.start = t - dt*0.5;
	twin.end = t + dt*0.5;
	if(gaps.find(twin)==gaps.end()) 
		return false;
	else
		return true;
}
// Forces samples in marked gaps to zero.  
void Time_Series::zero_gaps()
{
	double tsend;
	int i,istart,iend;
	set<Time_Window,Time_Window_Cmp>::iterator this_gap;
	tsend = t0+((double)(ns-1))*dt;

	for(this_gap=gaps.begin();this_gap!=gaps.end();++this_gap)
	{
		if(this_gap->end < t0) continue;
		if(this_gap->start > tsend) continue;
		if(this_gap->start<t0)
			istart = 0;
		else
			istart = nint((this_gap->start-t0)/dt);
		if(this_gap->end>tsend)
			iend = ns-1;
		else
			iend = nint((this_gap->end-t0)/dt);
		for(i=istart;i<=iend;++i) s[i]=0.0;
	}
}
// similar to above, but a gap in any component becomes a gap for
// the data.  The algorithm used here just hits all the members of the
// set called "gaps".  This should properly set all single channel gaps
// to zero in all three components.
void Three_Component_Seismogram::zero_gaps()
{
	double tsend;
	int i,istart,iend;
	set<Time_Window,Time_Window_Cmp>::iterator this_gap;
	tsend = t0+((double)(ns-1))*dt;

	for(this_gap=gaps.begin();this_gap!=gaps.end();++this_gap)
	{
		if(this_gap->end < t0) continue;
		if(this_gap->start > tsend) continue;
		if(this_gap->start<t0)
			istart = 0;
		else
			istart = nint((this_gap->start-t0)/dt);
		if(this_gap->end>tsend)
			iend = ns-1;
		else
			iend = nint((this_gap->end-t0)/dt);
		for(i=0;i<3;++i)
			for(int j=istart;j<=iend;++j)
				u(i,j)=0.0;
	}
}
} // Termination of namespace SEISPP definitions

